/***************************************************************
 * The code for AES-CTR with integer counter
 ***************************************************************/
#include"define.h"
#include<vector>

using namespace std;

class AESCTR {
public:
  __m128i w128[11]; // subkeys for AES128

  /***************************************************************
   * for debug
   ***************************************************************/
  void dispXMM(__m128i xmm) {

    MIE_ALIGN(128) unsigned char val[16] = { 0 };
    _mm_store_si128((__m128i*)(val), xmm);

    // little endian
    for (int i = 15; i >= 0; i--) {
      printf("%02X", val[i]);
    }
    printf("\n");
  }

  /***************************************************************
   * setup
   ***************************************************************/
  void aesKeySchedule(const unsigned char* k) {
    __m128i work = _mm_loadu_si128(reinterpret_cast<const __m128i*>(k));  // work = w3 : w2 : w1 : w0
    w128[0] = work;

    __m128i t, t2;

    #define EXPAND(n, RCON) \
        t  = _mm_slli_si128(work, 4);       /* t    = w2                : w1             : w0          : 0        */\
        t  = _mm_xor_si128(work, t);        /* t    = w3^w2             : w2^w1          : w1^w0       : w0       */\
        t2 = _mm_slli_si128(t, 8);          /* t2   = w1^w0             : w0             : 0           : 0        */\
        t  = _mm_xor_si128(t, t2);          /* t    = w3^w2^w1^w0       : w2^w1^w0       : w1^w0       : w0       */\
        work = _mm_aeskeygenassist_si128(work, RCON);           /* work = f(w3) : - : - : - */                      \
        work = _mm_shuffle_epi32(work, 0xFF);/*work = f(w3)             : f(w3)          : f(w3)       ; f(w3)    */\
        work = _mm_xor_si128(t, work);      /* work = w3^w2^w1^w0^f(w3) : w2^w1^w0^f(w3) : w1^w0^f(w3) : w0^f(w3) */\
        w128[n] = work;                     /* work = w7 : w6 : w5 : w4 */

    //
    EXPAND(1, 0x01);
    EXPAND(2, 0x02);
    EXPAND(3, 0x04);
    EXPAND(4, 0x08);
    EXPAND(5, 0x10);
    EXPAND(6, 0x20);
    EXPAND(7, 0x40);
    EXPAND(8, 0x80);
    EXPAND(9, 0x1b);
    EXPAND(10, 0x36);
  }

  void programSetUp(const unsigned char* k) {
    aesKeySchedule(k);  // run aes key scheulde and store subkey into w128
  }


  /***********************************************************************************
   * core
   ***********************************************************************************/
  inline void aes(__m128i st, __m128i& ks) {

    st = _mm_xor_si128(st, w128[0]);

    st = _mm_aesenc_si128(st, w128[1]);
    st = _mm_aesenc_si128(st, w128[2]);
    st = _mm_aesenc_si128(st, w128[3]);
    st = _mm_aesenc_si128(st, w128[4]);
    st = _mm_aesenc_si128(st, w128[5]);
    st = _mm_aesenc_si128(st, w128[6]);
    st = _mm_aesenc_si128(st, w128[7]);
    st = _mm_aesenc_si128(st, w128[8]);
    st = _mm_aesenc_si128(st, w128[9]);
    
    /*
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[1]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[2]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[3]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[4]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[5]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[6]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[7]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[8]) : );
    asm volatile("aesenc %[subkey], %[st]\n" : [st] "+x"(st) : [subkey] "x"(w128[9]) : );
    */
    ks = _mm_aesenclast_si128(st, w128[10]);
  }
  inline void aes_double(__m128i st0, __m128i st1, __m128i& ks0, __m128i& ks1) {

    st0 = _mm_xor_si128(st0, w128[0]);
    st1 = _mm_xor_si128(st1, w128[0]);

    st0 = _mm_aesenc_si128(st0, w128[1]);
    st1 = _mm_aesenc_si128(st1, w128[1]);

    st0 = _mm_aesenc_si128(st0, w128[2]);
    st1 = _mm_aesenc_si128(st1, w128[2]);
    
    st0 = _mm_aesenc_si128(st0, w128[3]);
    st1 = _mm_aesenc_si128(st1, w128[3]);
    
    st0 = _mm_aesenc_si128(st0, w128[4]);
    st1 = _mm_aesenc_si128(st1, w128[4]);
    
    st0 = _mm_aesenc_si128(st0, w128[5]);
    st1 = _mm_aesenc_si128(st1, w128[5]);
    
    st0 = _mm_aesenc_si128(st0, w128[6]);
    st1 = _mm_aesenc_si128(st1, w128[6]);
    
    st0 = _mm_aesenc_si128(st0, w128[7]);
    st1 = _mm_aesenc_si128(st1, w128[7]);
    
    st0 = _mm_aesenc_si128(st0, w128[8]);
    st1 = _mm_aesenc_si128(st1, w128[8]);
    
    st0 = _mm_aesenc_si128(st0, w128[9]);
    st1 = _mm_aesenc_si128(st1, w128[9]);

    ks0 = _mm_aesenclast_si128(st0, w128[10]);
    ks1 = _mm_aesenclast_si128(st1, w128[10]);
  }
  inline void aes_quad(__m128i st0, __m128i st1, __m128i st2, __m128i st3, __m128i& ks0, __m128i& ks1, __m128i& ks2, __m128i& ks3) {

    st0 = _mm_xor_si128(st0, w128[0]);
    st1 = _mm_xor_si128(st1, w128[0]);
    st2 = _mm_xor_si128(st2, w128[0]);
    st3 = _mm_xor_si128(st3, w128[0]);

    st0 = _mm_aesenc_si128(st0, w128[1]);
    st1 = _mm_aesenc_si128(st1, w128[1]);
    st2 = _mm_aesenc_si128(st2, w128[1]);
    st3 = _mm_aesenc_si128(st3, w128[1]);

    st0 = _mm_aesenc_si128(st0, w128[2]);
    st1 = _mm_aesenc_si128(st1, w128[2]);
    st2 = _mm_aesenc_si128(st2, w128[2]);
    st3 = _mm_aesenc_si128(st3, w128[2]);

    st0 = _mm_aesenc_si128(st0, w128[3]);
    st1 = _mm_aesenc_si128(st1, w128[3]);
    st2 = _mm_aesenc_si128(st2, w128[3]);
    st3 = _mm_aesenc_si128(st3, w128[3]);

    st0 = _mm_aesenc_si128(st0, w128[4]);
    st1 = _mm_aesenc_si128(st1, w128[4]);
    st2 = _mm_aesenc_si128(st2, w128[4]);
    st3 = _mm_aesenc_si128(st3, w128[4]);

    st0 = _mm_aesenc_si128(st0, w128[5]);
    st1 = _mm_aesenc_si128(st1, w128[5]);
    st2 = _mm_aesenc_si128(st2, w128[5]);
    st3 = _mm_aesenc_si128(st3, w128[5]);

    st0 = _mm_aesenc_si128(st0, w128[6]);
    st1 = _mm_aesenc_si128(st1, w128[6]);
    st2 = _mm_aesenc_si128(st2, w128[6]);
    st3 = _mm_aesenc_si128(st3, w128[6]);

    st0 = _mm_aesenc_si128(st0, w128[7]);
    st1 = _mm_aesenc_si128(st1, w128[7]);
    st2 = _mm_aesenc_si128(st2, w128[7]);
    st3 = _mm_aesenc_si128(st3, w128[7]);

    st0 = _mm_aesenc_si128(st0, w128[8]);
    st1 = _mm_aesenc_si128(st1, w128[8]);
    st2 = _mm_aesenc_si128(st2, w128[8]);
    st3 = _mm_aesenc_si128(st3, w128[8]);

    st0 = _mm_aesenc_si128(st0, w128[9]);
    st1 = _mm_aesenc_si128(st1, w128[9]);
    st2 = _mm_aesenc_si128(st2, w128[9]);
    st3 = _mm_aesenc_si128(st3, w128[9]);


    ks0 = _mm_aesenclast_si128(st0, w128[10]);
    ks1 = _mm_aesenclast_si128(st1, w128[10]);
    ks2 = _mm_aesenclast_si128(st2, w128[10]);
    ks3 = _mm_aesenclast_si128(st3, w128[10]);
  }
  inline void aes_sex(__m128i st0, __m128i st1, __m128i st2, __m128i st3, __m128i st4, __m128i st5, __m128i& ks0, __m128i& ks1, __m128i& ks2, __m128i& ks3, __m128i& ks4, __m128i& ks5) {

    st0 = _mm_xor_si128(st0, w128[0]);
    st1 = _mm_xor_si128(st1, w128[0]);
    st2 = _mm_xor_si128(st2, w128[0]);
    st3 = _mm_xor_si128(st3, w128[0]);
    st4 = _mm_xor_si128(st4, w128[0]);
    st5 = _mm_xor_si128(st5, w128[0]);

    st0 = _mm_aesenc_si128(st0, w128[1]);
    st1 = _mm_aesenc_si128(st1, w128[1]);
    st2 = _mm_aesenc_si128(st2, w128[1]);
    st3 = _mm_aesenc_si128(st3, w128[1]);
    st4 = _mm_aesenc_si128(st4, w128[1]);
    st5 = _mm_aesenc_si128(st5, w128[1]);

    st0 = _mm_aesenc_si128(st0, w128[2]);
    st1 = _mm_aesenc_si128(st1, w128[2]);
    st2 = _mm_aesenc_si128(st2, w128[2]);
    st3 = _mm_aesenc_si128(st3, w128[2]);
    st4 = _mm_aesenc_si128(st4, w128[2]);
    st5 = _mm_aesenc_si128(st5, w128[2]);

    st0 = _mm_aesenc_si128(st0, w128[3]);
    st1 = _mm_aesenc_si128(st1, w128[3]);
    st2 = _mm_aesenc_si128(st2, w128[3]);
    st3 = _mm_aesenc_si128(st3, w128[3]);
    st4 = _mm_aesenc_si128(st4, w128[3]);
    st5 = _mm_aesenc_si128(st5, w128[3]);
    
    st0 = _mm_aesenc_si128(st0, w128[4]);
    st1 = _mm_aesenc_si128(st1, w128[4]);
    st2 = _mm_aesenc_si128(st2, w128[4]);
    st3 = _mm_aesenc_si128(st3, w128[4]);
    st4 = _mm_aesenc_si128(st4, w128[4]);
    st5 = _mm_aesenc_si128(st5, w128[4]);
    
    st0 = _mm_aesenc_si128(st0, w128[5]);
    st1 = _mm_aesenc_si128(st1, w128[5]);
    st2 = _mm_aesenc_si128(st2, w128[5]);
    st3 = _mm_aesenc_si128(st3, w128[5]);
    st4 = _mm_aesenc_si128(st4, w128[5]);
    st5 = _mm_aesenc_si128(st5, w128[5]);
    
    st0 = _mm_aesenc_si128(st0, w128[6]);
    st1 = _mm_aesenc_si128(st1, w128[6]);
    st2 = _mm_aesenc_si128(st2, w128[6]);
    st3 = _mm_aesenc_si128(st3, w128[6]);
    st4 = _mm_aesenc_si128(st4, w128[6]);
    st5 = _mm_aesenc_si128(st5, w128[6]);
    
    st0 = _mm_aesenc_si128(st0, w128[7]);
    st1 = _mm_aesenc_si128(st1, w128[7]);
    st2 = _mm_aesenc_si128(st2, w128[7]);
    st3 = _mm_aesenc_si128(st3, w128[7]);
    st4 = _mm_aesenc_si128(st4, w128[7]);
    st5 = _mm_aesenc_si128(st5, w128[7]);
    
    st0 = _mm_aesenc_si128(st0, w128[8]);
    st1 = _mm_aesenc_si128(st1, w128[8]);
    st2 = _mm_aesenc_si128(st2, w128[8]);
    st3 = _mm_aesenc_si128(st3, w128[8]);
    st4 = _mm_aesenc_si128(st4, w128[8]);
    st5 = _mm_aesenc_si128(st5, w128[8]);
    
    st0 = _mm_aesenc_si128(st0, w128[9]);
    st1 = _mm_aesenc_si128(st1, w128[9]);
    st2 = _mm_aesenc_si128(st2, w128[9]);
    st3 = _mm_aesenc_si128(st3, w128[9]);
    st4 = _mm_aesenc_si128(st4, w128[9]);
    st5 = _mm_aesenc_si128(st5, w128[9]);


    ks0 = _mm_aesenclast_si128(st0, w128[10]);
    ks1 = _mm_aesenclast_si128(st1, w128[10]);
    ks2 = _mm_aesenclast_si128(st2, w128[10]);
    ks3 = _mm_aesenclast_si128(st3, w128[10]);
    ks4 = _mm_aesenclast_si128(st4, w128[10]);
    ks5 = _mm_aesenclast_si128(st5, w128[10]);
  }

};

AESCTR aesctr;

void aesctr_set_up(const unsigned char* k) {
  aesctr.programSetUp(k);
}

//
void aesctr_enc(
  const unsigned char* _K,     // Secret key                
  const unsigned char* _N,     // Nonce                     
  const unsigned char* _M,     // Message                   
  unsigned long long   _mlen,  // Length of Message         
  unsigned char* C_            // Ciphertext              
) {

  // set
  __m128i ctr0 = _mm_load_si128((const __m128i*)(_N));
  long long m_p = 0;
  __m128i ks0, ks1, ks2, ks3, ks4, ks5;
  __m128i add = _mm_cvtsi64_si128(1);

  
  while (m_p + 80 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_add_epi64(ctr0, add);
    __m128i ctr2 = _mm_add_epi64(ctr1, add);
    __m128i ctr3 = _mm_add_epi64(ctr2, add);
    __m128i ctr4 = _mm_add_epi64(ctr3, add);
    __m128i ctr5 = _mm_add_epi64(ctr4, add);

    // ks = enc(iv)
    aesctr.aes_sex(ctr0, ctr1, ctr2, ctr3, ctr4, ctr5, ks0, ks1, ks2, ks3, ks4, ks5);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    __m128i m2 = _mm_load_si128((const __m128i*)(_M + m_p));
    m2 = _mm_xor_si128(ks2, m2);
    _mm_store_si128((__m128i*)(C_ + m_p), m2);
    m_p += 16;

    __m128i m3 = _mm_load_si128((const __m128i*)(_M + m_p));
    m3 = _mm_xor_si128(ks3, m3);
    _mm_store_si128((__m128i*)(C_ + m_p), m3);
    m_p += 16;

    __m128i m4 = _mm_load_si128((const __m128i*)(_M + m_p));
    m4 = _mm_xor_si128(ks4, m4);
    _mm_store_si128((__m128i*)(C_ + m_p), m4);
    m_p += 16;

    __m128i m5 = _mm_load_si128((const __m128i*)(_M + m_p));
    m5 = _mm_xor_si128(ks5, m5);
    _mm_store_si128((__m128i*)(C_ + m_p), m5);
    m_p += 16;

    __m128i ctr0 = _mm_add_epi64(ctr5, add);
  }
  
  while (m_p + 48 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_add_epi64(ctr0, add);
    __m128i ctr2 = _mm_add_epi64(ctr1, add);
    __m128i ctr3 = _mm_add_epi64(ctr2, add);

    // ks = enc(iv)
    aesctr.aes_quad(ctr0, ctr1, ctr2, ctr3, ks0, ks1, ks2, ks3);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    __m128i m2 = _mm_load_si128((const __m128i*)(_M + m_p));
    m2 = _mm_xor_si128(ks2, m2);
    _mm_store_si128((__m128i*)(C_ + m_p), m2);
    m_p += 16;

    __m128i m3 = _mm_load_si128((const __m128i*)(_M + m_p));
    m3 = _mm_xor_si128(ks3, m3);
    _mm_store_si128((__m128i*)(C_ + m_p), m3);
    m_p += 16;

    ctr0 = _mm_add_epi64(ctr3, add);
  }
  
  while (m_p + 16 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_add_epi64(ctr0, add);

    // ks = enc(iv)
    aesctr.aes_double(ctr0, ctr1, ks0, ks1);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    ctr0 = _mm_add_epi64(ctr1, add);
  }

  while (m_p < (long long)(_mlen)) {

    // ks = enc(iv)
    aesctr.aes(ctr0, ks0);  

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    //ctr0 = _mm_add_epi64(ctr0, add);
  }




}
