/***************************************************************
 * The code for AES-PRF with Gray code counter
 ***************************************************************/
#include"define.h"
#include<vector>

using namespace std;

class AESPRF_GRAY {
public:
  __m128i w128[11]; // subkeys for AES128
  __m128i A[32];

  /***************************************************************
   * for debug
   ***************************************************************/
  void dispXMM(__m128i xmm) {

    MIE_ALIGN(128) unsigned char val[16] = { 0 };
    _mm_store_si128((__m128i*)(val), xmm);

    // little endian
    for (int i = 15; i >= 0; i--) {
      printf("%02X", val[i]);
    }
    printf("\n");
  }

  /***************************************************************
   * setup
   ***************************************************************/
  void aesKeySchedule(const unsigned char* k) {
    __m128i work = _mm_loadu_si128(reinterpret_cast<const __m128i*>(k));  // work = w3 : w2 : w1 : w0
    w128[0] = work;

    __m128i t, t2;

#define EXPAND(n, RCON) \
        t  = _mm_slli_si128(work, 4);       /* t    = w2                : w1             : w0          : 0        */\
        t  = _mm_xor_si128(work, t);        /* t    = w3^w2             : w2^w1          : w1^w0       : w0       */\
        t2 = _mm_slli_si128(t, 8);          /* t2   = w1^w0             : w0             : 0           : 0        */\
        t  = _mm_xor_si128(t, t2);          /* t    = w3^w2^w1^w0       : w2^w1^w0       : w1^w0       : w0       */\
        work = _mm_aeskeygenassist_si128(work, RCON);           /* work = f(w3) : - : - : - */                      \
        work = _mm_shuffle_epi32(work, 0xFF);/*work = f(w3)             : f(w3)          : f(w3)       ; f(w3)    */\
        work = _mm_xor_si128(t, work);      /* work = w3^w2^w1^w0^f(w3) : w2^w1^w0^f(w3) : w1^w0^f(w3) : w0^f(w3) */\
        w128[n] = work;                     /* work = w7 : w6 : w5 : w4 */

    //
    EXPAND(1, 0x01);
    EXPAND(2, 0x02);
    EXPAND(3, 0x04);
    EXPAND(4, 0x08);
    EXPAND(5, 0x10);
    EXPAND(6, 0x20);
    EXPAND(7, 0x40);
    EXPAND(8, 0x80);
    EXPAND(9, 0x1b);
    EXPAND(10, 0x36);
  }

  void programSetUp(const unsigned char* k) {
    aesKeySchedule(k);  // run aes key scheulde and store subkey into w128

    for (int i = 0; i < 32; i++) {

      __m128i x = _mm_cvtsi64_si128(1 << i);

      A[i] = x;
    }
  }


  /***********************************************************************************
   * core
   ***********************************************************************************/
  inline void aes(__m128i& st) {

    st = _mm_aesenc_si128(st, w128[1]);
    st = _mm_aesenc_si128(st, w128[2]);
    st = _mm_aesenc_si128(st, w128[3]);
    st = _mm_aesenc_si128(st, w128[4]);
    st = _mm_aesenc_si128(st, w128[5]);
    __m128i tmp = st;
    st = _mm_aesenc_si128(st, w128[6]);
    st = _mm_aesenc_si128(st, w128[7]);
    st = _mm_aesenc_si128(st, w128[8]);
    st = _mm_aesenc_si128(st, w128[9]);
    st = _mm_aesenclast_si128(st, w128[10]);

    st = _mm_xor_si128(st, tmp);
  }
  inline void aes_double(__m128i& st0, __m128i& st1) {

    st0 = _mm_aesenc_si128(st0, w128[1]);
    st1 = _mm_aesenc_si128(st1, w128[1]);

    st0 = _mm_aesenc_si128(st0, w128[2]);
    st1 = _mm_aesenc_si128(st1, w128[2]);

    st0 = _mm_aesenc_si128(st0, w128[3]);
    st1 = _mm_aesenc_si128(st1, w128[3]);

    st0 = _mm_aesenc_si128(st0, w128[4]);
    st1 = _mm_aesenc_si128(st1, w128[4]);

    st0 = _mm_aesenc_si128(st0, w128[5]);
    st1 = _mm_aesenc_si128(st1, w128[5]);

    __m128i tmp0 = st0;
    __m128i tmp1 = st1;

    st0 = _mm_aesenc_si128(st0, w128[6]);
    st1 = _mm_aesenc_si128(st1, w128[6]);

    st0 = _mm_aesenc_si128(st0, w128[7]);
    st1 = _mm_aesenc_si128(st1, w128[7]);

    st0 = _mm_aesenc_si128(st0, w128[8]);
    st1 = _mm_aesenc_si128(st1, w128[8]);

    st0 = _mm_aesenc_si128(st0, w128[9]);
    st1 = _mm_aesenc_si128(st1, w128[9]);

    st0 = _mm_aesenclast_si128(st0, w128[10]);
    st1 = _mm_aesenclast_si128(st1, w128[10]);

    st0 = _mm_xor_si128(st0, tmp0);
    st1 = _mm_xor_si128(st1, tmp1);

  }
  inline void aes_quad(__m128i& st0, __m128i& st1, __m128i& st2, __m128i& st3) {


    st0 = _mm_aesenc_si128(st0, w128[1]);
    st1 = _mm_aesenc_si128(st1, w128[1]);
    st2 = _mm_aesenc_si128(st2, w128[1]);
    st3 = _mm_aesenc_si128(st3, w128[1]);

    st0 = _mm_aesenc_si128(st0, w128[2]);
    st1 = _mm_aesenc_si128(st1, w128[2]);
    st2 = _mm_aesenc_si128(st2, w128[2]);
    st3 = _mm_aesenc_si128(st3, w128[2]);

    st0 = _mm_aesenc_si128(st0, w128[3]);
    st1 = _mm_aesenc_si128(st1, w128[3]);
    st2 = _mm_aesenc_si128(st2, w128[3]);
    st3 = _mm_aesenc_si128(st3, w128[3]);

    st0 = _mm_aesenc_si128(st0, w128[4]);
    st1 = _mm_aesenc_si128(st1, w128[4]);
    st2 = _mm_aesenc_si128(st2, w128[4]);
    st3 = _mm_aesenc_si128(st3, w128[4]);

    st0 = _mm_aesenc_si128(st0, w128[5]);
    st1 = _mm_aesenc_si128(st1, w128[5]);
    st2 = _mm_aesenc_si128(st2, w128[5]);
    st3 = _mm_aesenc_si128(st3, w128[5]);

    __m128i tmp0 = st0;
    __m128i tmp1 = st1;
    __m128i tmp2 = st2;
    __m128i tmp3 = st3;

    st0 = _mm_aesenc_si128(st0, w128[6]);
    st1 = _mm_aesenc_si128(st1, w128[6]);
    st2 = _mm_aesenc_si128(st2, w128[6]);
    st3 = _mm_aesenc_si128(st3, w128[6]);

    st0 = _mm_aesenc_si128(st0, w128[7]);
    st1 = _mm_aesenc_si128(st1, w128[7]);
    st2 = _mm_aesenc_si128(st2, w128[7]);
    st3 = _mm_aesenc_si128(st3, w128[7]);

    st0 = _mm_aesenc_si128(st0, w128[8]);
    st1 = _mm_aesenc_si128(st1, w128[8]);
    st2 = _mm_aesenc_si128(st2, w128[8]);
    st3 = _mm_aesenc_si128(st3, w128[8]);

    st0 = _mm_aesenc_si128(st0, w128[9]);
    st1 = _mm_aesenc_si128(st1, w128[9]);
    st2 = _mm_aesenc_si128(st2, w128[9]);
    st3 = _mm_aesenc_si128(st3, w128[9]);

    st0 = _mm_aesenclast_si128(st0, w128[10]);
    st1 = _mm_aesenclast_si128(st1, w128[10]);
    st2 = _mm_aesenclast_si128(st2, w128[10]);
    st3 = _mm_aesenclast_si128(st3, w128[10]);


    st0 = _mm_xor_si128(st0, tmp0);
    st1 = _mm_xor_si128(st1, tmp1);
    st2 = _mm_xor_si128(st2, tmp2);
    st3 = _mm_xor_si128(st3, tmp3);
  }
  inline void aes_oct(__m128i& st0, __m128i& st1, __m128i& st2, __m128i& st3, __m128i& st4, __m128i& st5, __m128i& st6, __m128i& st7) {

    st0 = _mm_aesenc_si128(st0, w128[1]);
    st1 = _mm_aesenc_si128(st1, w128[1]);
    st2 = _mm_aesenc_si128(st2, w128[1]);
    st3 = _mm_aesenc_si128(st3, w128[1]);
    st4 = _mm_aesenc_si128(st4, w128[1]);
    st5 = _mm_aesenc_si128(st5, w128[1]);
    st6 = _mm_aesenc_si128(st6, w128[1]);
    st7 = _mm_aesenc_si128(st7, w128[1]);

    st0 = _mm_aesenc_si128(st0, w128[2]);
    st1 = _mm_aesenc_si128(st1, w128[2]);
    st2 = _mm_aesenc_si128(st2, w128[2]);
    st3 = _mm_aesenc_si128(st3, w128[2]);
    st4 = _mm_aesenc_si128(st4, w128[2]);
    st5 = _mm_aesenc_si128(st5, w128[2]);
    st6 = _mm_aesenc_si128(st6, w128[2]);
    st7 = _mm_aesenc_si128(st7, w128[2]);

    st0 = _mm_aesenc_si128(st0, w128[3]);
    st1 = _mm_aesenc_si128(st1, w128[3]);
    st2 = _mm_aesenc_si128(st2, w128[3]);
    st3 = _mm_aesenc_si128(st3, w128[3]);
    st4 = _mm_aesenc_si128(st4, w128[3]);
    st5 = _mm_aesenc_si128(st5, w128[3]);
    st6 = _mm_aesenc_si128(st6, w128[3]);
    st7 = _mm_aesenc_si128(st7, w128[3]);

    st0 = _mm_aesenc_si128(st0, w128[4]);
    st1 = _mm_aesenc_si128(st1, w128[4]);
    st2 = _mm_aesenc_si128(st2, w128[4]);
    st3 = _mm_aesenc_si128(st3, w128[4]);
    st4 = _mm_aesenc_si128(st4, w128[4]);
    st5 = _mm_aesenc_si128(st5, w128[4]);
    st6 = _mm_aesenc_si128(st6, w128[4]);
    st7 = _mm_aesenc_si128(st7, w128[4]);

    st0 = _mm_aesenc_si128(st0, w128[5]);
    st1 = _mm_aesenc_si128(st1, w128[5]);
    st2 = _mm_aesenc_si128(st2, w128[5]);
    st3 = _mm_aesenc_si128(st3, w128[5]);
    st4 = _mm_aesenc_si128(st4, w128[5]);
    st5 = _mm_aesenc_si128(st5, w128[5]);
    st6 = _mm_aesenc_si128(st6, w128[5]);
    st7 = _mm_aesenc_si128(st7, w128[5]);

    __m128i tmp0 = st0;
    __m128i tmp1 = st1;
    __m128i tmp2 = st2;
    __m128i tmp3 = st3;
    __m128i tmp4 = st4;
    __m128i tmp5 = st5;
    __m128i tmp6 = st6;
    __m128i tmp7 = st7;

    st0 = _mm_aesenc_si128(st0, w128[6]);
    st1 = _mm_aesenc_si128(st1, w128[6]);
    st2 = _mm_aesenc_si128(st2, w128[6]);
    st3 = _mm_aesenc_si128(st3, w128[6]);
    st4 = _mm_aesenc_si128(st4, w128[6]);
    st5 = _mm_aesenc_si128(st5, w128[6]);
    st6 = _mm_aesenc_si128(st6, w128[6]);
    st7 = _mm_aesenc_si128(st7, w128[6]);

    st0 = _mm_aesenc_si128(st0, w128[7]);
    st1 = _mm_aesenc_si128(st1, w128[7]);
    st2 = _mm_aesenc_si128(st2, w128[7]);
    st3 = _mm_aesenc_si128(st3, w128[7]);
    st4 = _mm_aesenc_si128(st4, w128[7]);
    st5 = _mm_aesenc_si128(st5, w128[7]);
    st6 = _mm_aesenc_si128(st6, w128[7]);
    st7 = _mm_aesenc_si128(st7, w128[7]);

    st0 = _mm_aesenc_si128(st0, w128[8]);
    st1 = _mm_aesenc_si128(st1, w128[8]);
    st2 = _mm_aesenc_si128(st2, w128[8]);
    st3 = _mm_aesenc_si128(st3, w128[8]);
    st4 = _mm_aesenc_si128(st4, w128[8]);
    st5 = _mm_aesenc_si128(st5, w128[8]);
    st6 = _mm_aesenc_si128(st6, w128[8]);
    st7 = _mm_aesenc_si128(st7, w128[8]);

    st0 = _mm_aesenc_si128(st0, w128[9]);
    st1 = _mm_aesenc_si128(st1, w128[9]);
    st2 = _mm_aesenc_si128(st2, w128[9]);
    st3 = _mm_aesenc_si128(st3, w128[9]);
    st4 = _mm_aesenc_si128(st4, w128[9]);
    st5 = _mm_aesenc_si128(st5, w128[9]);
    st6 = _mm_aesenc_si128(st6, w128[9]);
    st7 = _mm_aesenc_si128(st7, w128[9]);

    st0 = _mm_aesenclast_si128(st0, w128[10]);
    st1 = _mm_aesenclast_si128(st1, w128[10]);
    st2 = _mm_aesenclast_si128(st2, w128[10]);
    st3 = _mm_aesenclast_si128(st3, w128[10]);
    st4 = _mm_aesenclast_si128(st4, w128[10]);
    st5 = _mm_aesenclast_si128(st5, w128[10]);
    st6 = _mm_aesenclast_si128(st6, w128[10]);
    st7 = _mm_aesenclast_si128(st7, w128[10]);


    st0 = _mm_xor_si128(st0, tmp0);
    st1 = _mm_xor_si128(st1, tmp1);
    st2 = _mm_xor_si128(st2, tmp2);
    st3 = _mm_xor_si128(st3, tmp3);
    st4 = _mm_xor_si128(st4, tmp4);
    st5 = _mm_xor_si128(st5, tmp5);
    st6 = _mm_xor_si128(st6, tmp6);
    st7 = _mm_xor_si128(st7, tmp7);
  }

};

AESPRF_GRAY aesprf_gray;

void aesprf_gray_set_up(const unsigned char* k) {
  aesprf_gray.programSetUp(k);
}

//
void aesprf_gray_enc(
  const unsigned char* _K,     // Secret key                
  const unsigned char* _N,     // Nonce                     
  const unsigned char* _M,     // Message                   
  unsigned long long   _mlen,  // Length of Message         
  unsigned char* C_            // Ciphertext              
) {

  // set
  __m128i ctr0 = _mm_load_si128((const __m128i*)(_N));
  ctr0 = _mm_xor_si128(ctr0, aesprf_gray.w128[0]);

  int cnt = 0;
  long long m_p = 0;

  while (m_p + 112 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_xor_si128(ctr0, aesprf_gray.A[0]);
    __m128i ctr2 = _mm_xor_si128(ctr1, aesprf_gray.A[1]);
    __m128i ctr3 = _mm_xor_si128(ctr2, aesprf_gray.A[0]);
    __m128i ctr4 = _mm_xor_si128(ctr3, aesprf_gray.A[2]);
    __m128i ctr5 = _mm_xor_si128(ctr4, aesprf_gray.A[0]);
    __m128i ctr6 = _mm_xor_si128(ctr5, aesprf_gray.A[1]);
    __m128i ctr7 = _mm_xor_si128(ctr6, aesprf_gray.A[0]);
    __m128i tmp = ctr7;

    // ks = enc(iv)
    aesprf_gray.aes_oct(ctr0, ctr1, ctr2, ctr3, ctr4, ctr5, ctr6, ctr7);

    //
    cnt += 8;
    int index = __builtin_ctz(cnt);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ctr0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ctr1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    __m128i m2 = _mm_load_si128((const __m128i*)(_M + m_p));
    m2 = _mm_xor_si128(ctr2, m2);
    _mm_store_si128((__m128i*)(C_ + m_p), m2);
    m_p += 16;

    __m128i m3 = _mm_load_si128((const __m128i*)(_M + m_p));
    m3 = _mm_xor_si128(ctr3, m3);
    _mm_store_si128((__m128i*)(C_ + m_p), m3);
    m_p += 16;

    __m128i m4 = _mm_load_si128((const __m128i*)(_M + m_p));
    m4 = _mm_xor_si128(ctr4, m4);
    _mm_store_si128((__m128i*)(C_ + m_p), m4);
    m_p += 16;

    __m128i m5 = _mm_load_si128((const __m128i*)(_M + m_p));
    m5 = _mm_xor_si128(ctr5, m5);
    _mm_store_si128((__m128i*)(C_ + m_p), m5);
    m_p += 16;

    __m128i m6 = _mm_load_si128((const __m128i*)(_M + m_p));
    m6 = _mm_xor_si128(ctr6, m6);
    _mm_store_si128((__m128i*)(C_ + m_p), m6);
    m_p += 16;

    __m128i m7 = _mm_load_si128((const __m128i*)(_M + m_p));
    m7 = _mm_xor_si128(ctr7, m7);
    _mm_store_si128((__m128i*)(C_ + m_p), m7);
    m_p += 16;

    ctr0 = _mm_xor_si128(tmp, aesprf_gray.A[index]);
  }

  while (m_p + 48 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_xor_si128(ctr0, aesprf_gray.A[0]);
    __m128i ctr2 = _mm_xor_si128(ctr1, aesprf_gray.A[1]);
    __m128i ctr3 = _mm_xor_si128(ctr2, aesprf_gray.A[0]);
    __m128i tmp = ctr3;

    // ks = enc(iv)
    aesprf_gray.aes_quad(ctr0, ctr1, ctr2, ctr3);

    //
    cnt += 4;
    //int index = __builtin_ctz(cnt);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ctr0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ctr1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    __m128i m2 = _mm_load_si128((const __m128i*)(_M + m_p));
    m2 = _mm_xor_si128(ctr2, m2);
    _mm_store_si128((__m128i*)(C_ + m_p), m2);
    m_p += 16;

    __m128i m3 = _mm_load_si128((const __m128i*)(_M + m_p));
    m3 = _mm_xor_si128(ctr3, m3);
    _mm_store_si128((__m128i*)(C_ + m_p), m3);
    m_p += 16;

    ctr0 = _mm_xor_si128(tmp, aesprf_gray.A[2]);
  }

  while (m_p + 16 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_xor_si128(ctr0, aesprf_gray.A[0]);
    __m128i tmp = ctr1;

    // ks = enc(iv)
    aesprf_gray.aes_double(ctr0, ctr1);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ctr0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ctr1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    ctr0 = _mm_xor_si128(tmp, aesprf_gray.A[1]);
  }

  while (m_p < (long long)(_mlen)) {

    // ks = enc(iv)
    aesprf_gray.aes(ctr0);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ctr0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    //ctr0 = _mm_add_epi64(ctr0, add);
  }




}
