#include"define.h"
#include<algorithm>

void speed_test_aesctr(int mlen) {

  //random number generator
  std::random_device rnd;
  std::mt19937_64 mt(rnd());

  // parameter
  unsigned char* m;
  unsigned char* c = NULL;
  unsigned char k[16] = { 0 };
  MIE_ALIGN(128) unsigned long long npub[2] = { 0 };

  // trialNum
  long long trialNum = 100000 *2048 * 2 / mlen;

  // assign
  m = (unsigned char*)malloc(mlen);       // assign memory to store plaintext
  c = (unsigned char*)malloc(mlen + 16);  // assign memory to store ciphertext

  // store measured data
  vector<double> box_insert;
  
  // repat 10 times
  int num_repeat = 10;
  double avg = 0;


  printf("========\n");
  printf("AES-CTR\n");
  printf("message length         = %d \n", mlen);
  printf("average cpb on %lld trials \n", trialNum);
  for (int repeat = 0; repeat < num_repeat; repeat++) {

    // random key
    for (int i = 0; i < 16; i++) {
      k[i] = mt() & 0xFF;
    }
    aesctr_set_up(k);

    //
    npub[0] = 0;
    for (int i = 1; i < 2; i++) {
      npub[i] = mt();                 // create random nonce
    }
    for (int i = 0; i < mlen; i++) {
      m[i] = mt() & 0xFF;             // create random data
    }

    // measure
    unsigned long long start, end;
    int CPUInfo[4];

    // start time
    __cpuid(CPUInfo, 0);
    start = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    for (int i = 0; i < trialNum; i++) {
      aesctr_enc(k, (unsigned char*)&npub, m, mlen, c);  // run
      npub[1]++;                                         // update nonce
    }

    // end time
    __cpuid(CPUInfo, 0);
    end = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    long long int totalcycles = (end - start);


    // print
    double cpb = (double)(totalcycles) / (mlen) / trialNum;
    printf("%.2f cpb  \n", cpb);
    box_insert.push_back(cpb);
  }
  sort(box_insert.begin(), box_insert.end());
  printf("\n");
  printf("%.2f cpb (min by %d repeat)\n", box_insert[0], num_repeat);
  printf("%.2f cpb (mid by %d repeat)\n", box_insert[num_repeat / 2 - 1], num_repeat);
  printf("%.2f cpb (max by %d repeat)\n", box_insert[num_repeat - 1], num_repeat);
  avg = 0;
  for (int i = 0; i < num_repeat; i++) {
    avg += box_insert[i];
  }
  printf("%.2f cpb (avg by %d repeat)\n", avg / num_repeat, num_repeat);

  free(m);
  free(c);


}
void speed_test_aesprf(int mlen) {

  //random
  std::random_device rnd;
  std::mt19937_64 mt(rnd());

  // parameter
  unsigned char* m;
  unsigned char* c = NULL;
  unsigned char k[16] = { 0 };
  MIE_ALIGN(128) unsigned long long npub[2] = { 0 };

  long long trialNum = 100000 * 2048 * 2 / mlen;

  // assign
  m = (unsigned char*)malloc(mlen);       // assign memory to store plaintext
  c = (unsigned char*)malloc(mlen + 16);  // assign memory to store ciphertext

  vector<double> box_insert;

  int num_repeat = 10;
  double avg = 0;


  printf("========\n");
  printf("AES-PRF\n");
  printf("message length         = %d \n", mlen);
  printf("average cpb on %lld trials \n", trialNum);
  for (int repeat = 0; repeat < num_repeat; repeat++) {

    // random key
    for (int i = 0; i < 16; i++) {
      k[i] = mt() & 0xFF;
    }
    aesprf_set_up(k);

    //
    npub[0] = 0;
    for (int i = 1; i < 2; i++) {
      npub[i] = mt();          // create random nonce
    }
    for (int i = 0; i < mlen; i++) {
      m[i] = mt() & 0xFF;             // create random data
    }

    // measure
    unsigned long long start, end;
    int CPUInfo[4];

    // start time
    __cpuid(CPUInfo, 0);
    start = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    for (int i = 0; i < trialNum; i++) {
      aesprf_enc(k, (unsigned char*)&npub, m, mlen, c); // run
      npub[1]++;                                                    // update nonce
    }

    // end time
    __cpuid(CPUInfo, 0);
    end = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    long long int totalcycles = (end - start);


    // print
    double cpb = (double)(totalcycles) / (mlen) / trialNum;
    printf("%.2f cpb  \n", cpb);
    box_insert.push_back(cpb);
  }
  sort(box_insert.begin(), box_insert.end());
  printf("\n");
  printf("%.2f cpb (min by %d repeat)\n", box_insert[0], num_repeat);
  printf("%.2f cpb (mid by %d repeat)\n", box_insert[num_repeat / 2 - 1], num_repeat);
  printf("%.2f cpb (max by %d repeat)\n", box_insert[num_repeat - 1], num_repeat);
  avg = 0;
  for (int i = 0; i < num_repeat; i++) {
    avg += box_insert[i];
  }
  printf("%.2f cpb (avg by %d repeat)\n", avg / num_repeat, num_repeat);

  free(m);
  free(c);


}
void speed_test_zipaes(int mlen) {

  //random
  std::random_device rnd;
  std::mt19937_64 mt(rnd());

  // parameter
  unsigned char* m;
  unsigned char* c = NULL;
  unsigned char k[16] = { 0 };
  MIE_ALIGN(128) unsigned long long npub[2] = { 0 };

  long long trialNum = 100000 * 2048 * 2 / mlen;

  // assign
  m = (unsigned char*)malloc(mlen);  // assign memory to store plaintext
  c = (unsigned char*)malloc(mlen + 16);  // assign memory to store ciphertext

  vector<double> box_insert;

  int num_repeat = 10;
  double avg = 0;


  printf("========\n");
  printf("ZIP-AES\n");
  printf("message length         = %d \n", mlen);
  printf("average cpb on %lld trials \n", trialNum);
  for (int repeat = 0; repeat < num_repeat; repeat++) {

    // random key
    for (int i = 0; i < 16; i++) {
      k[i] = mt() & 0xFF;
    }
    zipaes_set_up(k);

    //
    npub[0] = 0;
    for (int i = 1; i < 2; i++) {
      npub[i] = mt();          // create random nonce
    }
    for (int i = 0; i < mlen; i++) {
      m[i] = mt() & 0xFF;             // create random data
    }

    // measure
    unsigned long long start, end;
    int CPUInfo[4];

    // start time
    __cpuid(CPUInfo, 0);
    start = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    for (int i = 0; i < trialNum; i++) {
      zipaes_enc(k, (unsigned char*)&npub, m, mlen, c); // run
      npub[1]++;                                                    // update nonce
    }

    // end time
    __cpuid(CPUInfo, 0);
    end = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    long long int totalcycles = (end - start);


    // print
    double cpb = (double)(totalcycles) / (mlen) / trialNum;
    printf("%.2f cpb  \n", cpb);
    box_insert.push_back(cpb);
  }
  sort(box_insert.begin(), box_insert.end());
  printf("\n");
  printf("%.2f cpb (min by %d repeat)\n", box_insert[0], num_repeat);
  printf("%.2f cpb (mid by %d repeat)\n", box_insert[num_repeat / 2 - 1], num_repeat);
  printf("%.2f cpb (max by %d repeat)\n", box_insert[num_repeat - 1], num_repeat);
  avg = 0;
  for (int i = 0; i < num_repeat; i++) {
    avg += box_insert[i];
  }
  printf("%.2f cpb (avg by %d repeat)\n", avg / num_repeat, num_repeat);

  free(m);
  free(c);


}
void speed_test_aesctr_gray(int mlen) {

  //random
  std::random_device rnd;
  std::mt19937_64 mt(rnd());

  // parameter
  unsigned char* m;
  unsigned char* c = NULL;
  unsigned char k[16] = { 0 };
  MIE_ALIGN(128) unsigned long long npub[2] = { 0 };

  //
  long long trialNum = 100000 * 2048 * 2 / mlen;

  // assign
  m = (unsigned char*)malloc(mlen);  // assign memory to store plaintext
  c = (unsigned char*)malloc(mlen + 16);  // assign memory to store ciphertext

  vector<double> box_insert;

  int num_repeat = 10;
  double avg = 0;


  printf("========\n");
  printf("AES-CTR (gray)\n");
  printf("message length         = %d \n", mlen);
  printf("average cpb on %lld trials \n", trialNum);
  for (int repeat = 0; repeat < num_repeat; repeat++) {

    // random key
    for (int i = 0; i < 16; i++) {
      k[i] = mt() & 0xFF;
    }
    aesctr_gray_set_up(k);

    //
    npub[0] = 0;
    for (int i = 1; i < 2; i++) {
      npub[i] = mt();          // create random nonce
    }
    for (int i = 0; i < mlen; i++) {
      m[i] = mt() & 0xFF;             // create random data
    }

    // measure
    unsigned long long start, end;
    int CPUInfo[4];

    // start time
    __cpuid(CPUInfo, 0);
    start = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    for (int i = 0; i < trialNum; i++) {
      aesctr_gray_enc(k, (unsigned char*)&npub, m, mlen, c); // run
      npub[1]++;                                                    // update nonce
    }

    // end time
    __cpuid(CPUInfo, 0);
    end = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    long long int totalcycles = (end - start);


    // print
    double cpb = (double)(totalcycles) / (mlen) / trialNum;
    printf("%.2f cpb  \n", cpb);
    box_insert.push_back(cpb);
  }
  sort(box_insert.begin(), box_insert.end());
  printf("\n");
  printf("%.2f cpb (min by %d repeat)\n", box_insert[0], num_repeat);
  printf("%.2f cpb (mid by %d repeat)\n", box_insert[num_repeat / 2 - 1], num_repeat);
  printf("%.2f cpb (max by %d repeat)\n", box_insert[num_repeat - 1], num_repeat);
  avg = 0;
  for (int i = 0; i < num_repeat; i++) {
    avg += box_insert[i];
  }
  printf("%.2f cpb (avg by %d repeat)\n", avg / num_repeat, num_repeat);

  free(m);
  free(c);


}
void speed_test_aesprf_gray(int mlen) {

  //random
  std::random_device rnd;
  std::mt19937_64 mt(rnd());

  // parameter
  unsigned char* m;
  unsigned char* c = NULL;
  unsigned char k[16] = { 0 };
  MIE_ALIGN(128) unsigned long long npub[2] = { 0 };

  //
  long long trialNum = 100000 * 2048 * 2 / mlen;

  // assign
  m = (unsigned char*)malloc(mlen);  // assign memory to store plaintext
  c = (unsigned char*)malloc(mlen + 16);  // assign memory to store ciphertext

  vector<double> box_insert;

  int num_repeat = 10;
  double avg = 0;


  printf("========\n");
  printf("AES-PRF (gray)\n");
  printf("message length         = %d \n", mlen);
  printf("average cpb on %lld trials \n", trialNum);
  for (int repeat = 0; repeat < num_repeat; repeat++) {

    // random key
    for (int i = 0; i < 16; i++) {
      k[i] = mt() & 0xFF;
    }
    aesprf_gray_set_up(k);

    //
    npub[0] = 0;
    for (int i = 1; i < 2; i++) {
      npub[i] = mt();          // create random nonce
    }
    for (int i = 0; i < mlen; i++) {
      m[i] = mt() & 0xFF;             // create random data
    }

    // measure
    unsigned long long start, end;
    int CPUInfo[4];

    // start time
    __cpuid(CPUInfo, 0);
    start = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    for (int i = 0; i < trialNum; i++) {
      aesprf_gray_enc(k, (unsigned char*)&npub, m, mlen, c); // run
      npub[1]++;                                                    // update nonce
    }

    // end time
    __cpuid(CPUInfo, 0);
    end = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    long long int totalcycles = (end - start);


    // print
    double cpb = (double)(totalcycles) / (mlen) / trialNum;
    printf("%.2f cpb  \n", cpb);
    box_insert.push_back(cpb);
  }
  sort(box_insert.begin(), box_insert.end());
  printf("\n");
  printf("%.2f cpb (min by %d repeat)\n", box_insert[0], num_repeat);
  printf("%.2f cpb (mid by %d repeat)\n", box_insert[num_repeat / 2 - 1], num_repeat);
  printf("%.2f cpb (max by %d repeat)\n", box_insert[num_repeat - 1], num_repeat);
  avg = 0;
  for (int i = 0; i < num_repeat; i++) {
    avg += box_insert[i];
  }
  printf("%.2f cpb (avg by %d repeat)\n", avg / num_repeat, num_repeat);

  free(m);
  free(c);


}
void speed_test_zipaes_gray(int mlen) {

  //random
  std::random_device rnd;
  std::mt19937_64 mt(rnd());

  // parameter
  unsigned char* m;
  unsigned char* c = NULL;
  unsigned char k[16] = { 0 };
  MIE_ALIGN(128) unsigned long long npub[2] = { 0 };

  //
  long long trialNum = 100000 * 2048 * 2 / mlen;

  // assign
  m = (unsigned char*)malloc(mlen);  // assign memory to store plaintext
  c = (unsigned char*)malloc(mlen + 16);  // assign memory to store ciphertext

  vector<double> box_insert;

  int num_repeat = 10;
  double avg = 0;


  printf("========\n");
  printf("ZIP-AES (gray)\n");
  printf("message length         = %d \n", mlen);
  printf("average cpb on %lld trials \n", trialNum);
  for (int repeat = 0; repeat < num_repeat; repeat++) {

    // random key
    for (int i = 0; i < 16; i++) {
      k[i] = mt() & 0xFF;
    }
    zipaes_gray_set_up(k);

    //
    npub[0] = 0;
    for (int i = 1; i < 2; i++) {
      npub[i] = mt();          // create random nonce
    }
    for (int i = 0; i < mlen; i++) {
      m[i] = mt() & 0xFF;             // create random data
    }

    // measure
    unsigned long long start, end;
    int CPUInfo[4];

    // start time
    __cpuid(CPUInfo, 0);
    start = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    for (int i = 0; i < trialNum; i++) {
      zipaes_gray_enc(k, (unsigned char*)&npub, m, mlen, c); // run
      npub[1]++;                                                    // update nonce
    }

    // end time
    __cpuid(CPUInfo, 0);
    end = __rdtsc();
    __cpuid(CPUInfo, 0);

    //
    long long int totalcycles = (end - start);


    // print
    double cpb = (double)(totalcycles) / (mlen) / trialNum;
    printf("%.2f cpb  \n", cpb);
    box_insert.push_back(cpb);
  }
  sort(box_insert.begin(), box_insert.end());
  printf("\n");
  printf("%.2f cpb (min by %d repeat)\n", box_insert[0], num_repeat);
  printf("%.2f cpb (mid by %d repeat)\n", box_insert[num_repeat / 2 - 1], num_repeat);
  printf("%.2f cpb (max by %d repeat)\n", box_insert[num_repeat - 1], num_repeat);
  avg = 0;
  for (int i = 0; i < num_repeat; i++) {
    avg += box_insert[i];
  }
  printf("%.2f cpb (avg by %d repeat)\n", avg / num_repeat, num_repeat);

  free(m);
  free(c);


}

void experiment_autocorrelation(void) {

  unsigned char tmp[16] = { 0x01 ,0x23 ,0x45 ,0x67 ,0x89 ,0xAB ,0xCD ,0xEF ,0xFE ,0xDC ,0xBA ,0x98 ,0x76 ,0x54 ,0x32 ,0x10 };
  srand(time(NULL));

  for (int i = 0; i < 16; i++) {
    tmp[i] = rand() % 0xFF;
  }

  zipaes_set_up(tmp);
  double cor = experimentDifferentialLinear();
}
void experiment_differential_and_linear(void) {

  unsigned char tmp[16] = { 0x01 ,0x23 ,0x45 ,0x67 ,0x89 ,0xAB ,0xCD ,0xEF ,0xFE ,0xDC ,0xBA ,0x98 ,0x76 ,0x54 ,0x32 ,0x10 };
  srand(time(NULL));


  double avgRight = 0;
  double avgWrong = 0;
  for (int loop = 0; loop < 10; loop++) {

    for (int i = 0; i < 16; i++) {
      tmp[i] = rand() % 0xFF;
    }

    zipaes_set_up(tmp);


    cout << "right guess" << endl;
    avgRight += abs(experimentDifferentialAndLinear(1));
    cout << endl;

    cout << "wrong (random) guess" << endl;
    avgWrong += experimentDifferentialAndLinear(0);
    cout << endl;

  }

  cout << "average for 10 right keys : " << log2(avgRight / 10) << endl;
  cout << "average for 10 wrong keys : " << log2(avgWrong / 10) << endl;

}


int main(int argc, char *argv[]) {

  unsigned long long mlen = 0;

  int aesctr = 0;
  int aesprf = 0;
  int zipaes = 0;
  int gray = 0;

  int exauto = 0;
  int exdifflin = 0;
  
  for (int i = 0; i < argc; i++) {
    if (!strcmp(argv[i], "-aesctr")) {
      aesctr = 1;
    }
    if (!strcmp(argv[i], "-aesprf")) {
      aesprf = 1;
    }
    if (!strcmp(argv[i], "-zipaes")) {
      zipaes = 1;
    }
    if (!strcmp(argv[i], "-gray")) {
      gray = 1;
    }

    if (!strcmp(argv[i], "-mlen")) {
      mlen = atoi(argv[i + 1]);
    }


    if (!strcmp(argv[i], "-exauto")) {
      exauto = 1;
    }


    if (!strcmp(argv[i], "-exdifflin")) {
      exdifflin = 1;
    }
  }

  if (exauto) {
    cout << "Experiments for autocorrelation" << endl;
    experiment_autocorrelation();
    return 0;
  }


  if (exdifflin) {
    cout << "Experiments for differential-and-linear key recovery" << endl;
    experiment_differential_and_linear();
    return 0;
  }

  



  if (gray) {
    if (aesctr) {

      if (mlen == 0) {
        speed_test_aesctr_gray(131072);
        speed_test_aesctr_gray(16384);
        speed_test_aesctr_gray(2048);
        speed_test_aesctr_gray(256);
        speed_test_aesctr_gray(32);
        speed_test_aesctr_gray(16);
      }
      else {
        speed_test_aesctr_gray(mlen);
      }

    }
    else if (aesprf) {

      if (mlen == 0) {
        speed_test_aesprf_gray(131072);
        speed_test_aesprf_gray(16384);
        speed_test_aesprf_gray(2048);
        speed_test_aesprf_gray(256);
        speed_test_aesprf_gray(32);
        speed_test_aesprf_gray(16);
      }
      else {
        speed_test_aesprf_gray(mlen);
      }
    }
    else if (zipaes) {

      if (mlen == 0) {
        speed_test_zipaes_gray(131072);
        speed_test_zipaes_gray(16384);
        speed_test_zipaes_gray(2048);
        speed_test_zipaes_gray(256);
        speed_test_zipaes_gray(32);
        speed_test_zipaes_gray(16);
      }
      else {
        speed_test_zipaes_gray(mlen);
      }
    }

  }
  else {
    if (aesctr) {

      if (mlen == 0) {
        speed_test_aesctr(131072);
        speed_test_aesctr(16384);
        speed_test_aesctr(2048);
        speed_test_aesctr(256);
        speed_test_aesctr(32);
        speed_test_aesctr(16);
      }
      else {
        speed_test_aesctr(mlen);
      }

    }
    else if (aesprf) {

      if (mlen == 0) {
        speed_test_aesprf(131072);
        speed_test_aesprf(16384);
        speed_test_aesprf(2048);
        speed_test_aesprf(256);
        speed_test_aesprf(32);
        speed_test_aesprf(16);
      }
      else {
        speed_test_aesprf(mlen);
      }
    }
    else if (zipaes) {

      if (mlen == 0) {
        speed_test_zipaes(131072);
        speed_test_zipaes(16384);
        speed_test_zipaes(2048);
        speed_test_zipaes(256);
        speed_test_zipaes(32);
        speed_test_zipaes(16);
      }
      else {
        speed_test_zipaes(mlen);
      }
    }

  }



  return 0;



}