/***************************************************************
 * The code for ZIP-AES with integer counter
 * The code of experiments for differential-linear (autocorrelation)
 * The code of experiments for differential-and-linear key recovery
 ***************************************************************/
#include"define.h"
#include<vector>

using namespace std;

class ZIPAES {
public:
  __m128i w128[11]; // subkeys for AES128
  __m128i zero; 

  /***************************************************************
   * for debug
   ***************************************************************/
  void dispXMM(__m128i xmm) {

    MIE_ALIGN(128) unsigned char val[16] = { 0 };
    _mm_store_si128((__m128i*)(val), xmm);

    // little endian
    for (int i = 15; i >= 0; i--) {
      printf("%02X", val[i]);
    }
    printf("\n");
  }

  /***************************************************************
   * setup
   ***************************************************************/
  void aesKeySchedule(const unsigned char* k) {
    __m128i work = _mm_loadu_si128(reinterpret_cast<const __m128i*>(k));  // work = w3 : w2 : w1 : w0
    w128[0] = work;

    __m128i t, t2;

#define EXPAND(n, RCON) \
        t  = _mm_slli_si128(work, 4);       /* t    = w2                : w1             : w0          : 0        */\
        t  = _mm_xor_si128(work, t);        /* t    = w3^w2             : w2^w1          : w1^w0       : w0       */\
        t2 = _mm_slli_si128(t, 8);          /* t2   = w1^w0             : w0             : 0           : 0        */\
        t  = _mm_xor_si128(t, t2);          /* t    = w3^w2^w1^w0       : w2^w1^w0       : w1^w0       : w0       */\
        work = _mm_aeskeygenassist_si128(work, RCON);           /* work = f(w3) : - : - : - */                      \
        work = _mm_shuffle_epi32(work, 0xFF);/*work = f(w3)             : f(w3)          : f(w3)       ; f(w3)    */\
        work = _mm_xor_si128(t, work);      /* work = w3^w2^w1^w0^f(w3) : w2^w1^w0^f(w3) : w1^w0^f(w3) : w0^f(w3) */\
        w128[n] = work;                     /* work = w7 : w6 : w5 : w4 */

    //
    EXPAND(1, 0x01);
    EXPAND(2, 0x02);
    EXPAND(3, 0x04);
    EXPAND(4, 0x08);
    EXPAND(5, 0x10);
    EXPAND(6, 0x20);
    EXPAND(7, 0x40);
    EXPAND(8, 0x80);
    EXPAND(9, 0x1b);
    EXPAND(10, 0x36);


    //w128[6] = _mm_aesimc_si128(w128[6]);
    w128[7] = _mm_aesimc_si128(w128[7]);
    w128[8] = _mm_aesimc_si128(w128[8]);
    w128[9] = _mm_aesimc_si128(w128[9]);
    w128[10] = _mm_aesimc_si128(w128[10]);
  }
  void programSetUp(const unsigned char* k) {
    aesKeySchedule(k);  // run aes key scheulde and store subkey into w128
    zero = _mm_setzero_si128();
  }


  /***********************************************************************************
   * core
   ***********************************************************************************/
  inline void aes(__m128i st, __m128i& ks) {

    __m128i stL = _mm_xor_si128(st, w128[0]);
    __m128i stR = _mm_xor_si128(st, w128[6]);
    stR = _mm_aesimc_si128(stR);

    stL = _mm_aesenc_si128(stL, w128[1]);
    stR = _mm_aesdec_si128(stR, w128[7]);

    stL = _mm_aesenc_si128(stL, w128[2]);
    stR = _mm_aesdec_si128(stR, w128[8]);

    stL = _mm_aesenc_si128(stL, w128[3]);
    stR = _mm_aesdec_si128(stR, w128[9]);

    stL = _mm_aesenc_si128(stL, w128[4]);
    stR = _mm_aesdec_si128(stR, w128[10]);

    stL = _mm_aesenc_si128(stL, w128[5]);
    ks = _mm_aesdeclast_si128(stR, stL);

  }
  inline void aes_double(__m128i st0, __m128i st1, __m128i& ks0, __m128i& ks1) {

    __m128i st0L = _mm_xor_si128(st0, w128[0]);
    __m128i st1L = _mm_xor_si128(st1, w128[0]);

    __m128i st0R = _mm_xor_si128(st0, w128[6]);
    __m128i st1R = _mm_xor_si128(st1, w128[6]);

    st0R = _mm_aesimc_si128(st0R);
    st1R = _mm_aesimc_si128(st1R);


    st0L = _mm_aesenc_si128(st0L, w128[1]);
    st1L = _mm_aesenc_si128(st1L, w128[1]);

    st0R = _mm_aesdec_si128(st0R, w128[7]);
    st1R = _mm_aesdec_si128(st1R, w128[7]);

    st0L = _mm_aesenc_si128(st0L, w128[2]);
    st1L = _mm_aesenc_si128(st1L, w128[2]);

    st0R = _mm_aesdec_si128(st0R, w128[8]);
    st1R = _mm_aesdec_si128(st1R, w128[8]);

    st0L = _mm_aesenc_si128(st0L, w128[3]);
    st1L = _mm_aesenc_si128(st1L, w128[3]);

    st0R = _mm_aesdec_si128(st0R, w128[9]);
    st1R = _mm_aesdec_si128(st1R, w128[9]);

    st0L = _mm_aesenc_si128(st0L, w128[4]);
    st1L = _mm_aesenc_si128(st1L, w128[4]);

    st0R = _mm_aesdec_si128(st0R, w128[10]);
    st1R = _mm_aesdec_si128(st1R, w128[10]);

    st0L = _mm_aesenc_si128(st0L, w128[5]);
    st1L = _mm_aesenc_si128(st1L, w128[5]);

    ks0 = _mm_aesdeclast_si128(st0R, st0L);
    ks1 = _mm_aesdeclast_si128(st1R, st1L);


  }
  inline void aes_quad(__m128i st0, __m128i st1, __m128i st2, __m128i st3, __m128i& ks0, __m128i& ks1, __m128i& ks2, __m128i& ks3) {

    __m128i st0L = _mm_xor_si128(st0, w128[0]);
    __m128i st1L = _mm_xor_si128(st1, w128[0]);
    __m128i st2L = _mm_xor_si128(st2, w128[0]);
    __m128i st3L = _mm_xor_si128(st3, w128[0]);

    __m128i st0R = _mm_xor_si128(st0, w128[6]);
    __m128i st1R = _mm_xor_si128(st1, w128[6]);
    __m128i st2R = _mm_xor_si128(st2, w128[6]);
    __m128i st3R = _mm_xor_si128(st3, w128[6]);

    st0R = _mm_aesimc_si128(st0R);
    st1R = _mm_aesimc_si128(st1R);
    st2R = _mm_aesimc_si128(st2R);
    st3R = _mm_aesimc_si128(st3R);

    st0L = _mm_aesenc_si128(st0L, w128[1]);
    st1L = _mm_aesenc_si128(st1L, w128[1]);
    st2L = _mm_aesenc_si128(st2L, w128[1]);
    st3L = _mm_aesenc_si128(st3L, w128[1]);

    st0R = _mm_aesdec_si128(st0R, w128[7]);
    st1R = _mm_aesdec_si128(st1R, w128[7]);
    st2R = _mm_aesdec_si128(st2R, w128[7]);
    st3R = _mm_aesdec_si128(st3R, w128[7]);

    st0L = _mm_aesenc_si128(st0L, w128[2]);
    st1L = _mm_aesenc_si128(st1L, w128[2]);
    st2L = _mm_aesenc_si128(st2L, w128[2]);
    st3L = _mm_aesenc_si128(st3L, w128[2]);

    st0R = _mm_aesdec_si128(st0R, w128[8]);
    st1R = _mm_aesdec_si128(st1R, w128[8]);
    st2R = _mm_aesdec_si128(st2R, w128[8]);
    st3R = _mm_aesdec_si128(st3R, w128[8]);

    st0L = _mm_aesenc_si128(st0L, w128[3]);
    st1L = _mm_aesenc_si128(st1L, w128[3]);
    st2L = _mm_aesenc_si128(st2L, w128[3]);
    st3L = _mm_aesenc_si128(st3L, w128[3]);

    st0R = _mm_aesdec_si128(st0R, w128[9]);
    st1R = _mm_aesdec_si128(st1R, w128[9]);
    st2R = _mm_aesdec_si128(st2R, w128[9]);
    st3R = _mm_aesdec_si128(st3R, w128[9]);

    st0L = _mm_aesenc_si128(st0L, w128[4]);
    st1L = _mm_aesenc_si128(st1L, w128[4]);
    st2L = _mm_aesenc_si128(st2L, w128[4]);
    st3L = _mm_aesenc_si128(st3L, w128[4]);

    st0R = _mm_aesdec_si128(st0R, w128[10]);
    st1R = _mm_aesdec_si128(st1R, w128[10]);
    st2R = _mm_aesdec_si128(st2R, w128[10]);
    st3R = _mm_aesdec_si128(st3R, w128[10]);

    st0L = _mm_aesenc_si128(st0L, w128[5]);
    st1L = _mm_aesenc_si128(st1L, w128[5]);
    st2L = _mm_aesenc_si128(st2L, w128[5]);
    st3L = _mm_aesenc_si128(st3L, w128[5]);

    ks0 = _mm_aesdeclast_si128(st0R, st0L);
    ks1 = _mm_aesdeclast_si128(st1R, st1L);
    ks2 = _mm_aesdeclast_si128(st2R, st2L);
    ks3 = _mm_aesdeclast_si128(st3R, st3L);

  }


  /***********************************************************************************
   * experimental code
   ***********************************************************************************/
  inline void aes(__m128i st, __m128i& ks, int round) {

    __m128i stL = _mm_xor_si128(st, w128[0]);
    __m128i stR = _mm_xor_si128(st, w128[6]);
    stR = _mm_aesimc_si128(stR);

    for (int i = 0; i < round - 1; i++) {
      stL = _mm_aesenc_si128(stL, w128[1 + i]);
      stR = _mm_aesdec_si128(stR, w128[7 + i]);
    }
    stL = _mm_aesenc_si128(stL, w128[1 + round - 1]);
    ks = _mm_aesdeclast_si128(stR, stL);

  }

  inline void aesL(__m128i st, __m128i& ks, int round) {

    __m128i stL = _mm_xor_si128(st, w128[0]);
    for (int r = 0; r < round; r++) {
      stL = _mm_aesenc_si128(stL, w128[r + 1]);
    }
    ks = stL;
  }
  inline void aesR(__m128i st, __m128i& ks, int round) {

    __m128i stR = _mm_xor_si128(st, w128[6]);
    stR = _mm_aesimc_si128(stR);

    for (int r = 0; r < round - 1; r++) {
      stR = _mm_aesdec_si128(stR, w128[7 + r]);
    }
    stR = _mm_aesdeclast_si128(stR, _mm_setzero_si128());

    ks = stR;
  }

};

ZIPAES zipaes;

void zipaes_set_up(const unsigned char* k) {
  zipaes.programSetUp(k);
}

//
void zipaes_enc(
  const unsigned char* _K,     // Secret key                
  const unsigned char* _N,     // Nonce                     
  const unsigned char* _M,     // Message                   
  unsigned long long   _mlen,  // Length of Message         
  unsigned char* C_            // Ciphertext              
) {

  // set
  __m128i ctr0 = _mm_load_si128((const __m128i*)(_N));
  long long m_p = 0;
  __m128i ks0, ks1, ks2, ks3, ks4, ks5;
  __m128i add = _mm_cvtsi64_si128(1);

  
  while (m_p + 48 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_add_epi64(ctr0, add);
    __m128i ctr2 = _mm_add_epi64(ctr1, add);
    __m128i ctr3 = _mm_add_epi64(ctr2, add);

    // ks = enc(iv)
    zipaes.aes_quad(ctr0, ctr1, ctr2, ctr3, ks0, ks1, ks2, ks3);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    __m128i m2 = _mm_load_si128((const __m128i*)(_M + m_p));
    m2 = _mm_xor_si128(ks2, m2);
    _mm_store_si128((__m128i*)(C_ + m_p), m2);
    m_p += 16;

    __m128i m3 = _mm_load_si128((const __m128i*)(_M + m_p));
    m3 = _mm_xor_si128(ks3, m3);
    _mm_store_si128((__m128i*)(C_ + m_p), m3);
    m_p += 16;

    ctr0 = _mm_add_epi64(ctr3, add);
  }
  
  while (m_p + 16 < (long long)(_mlen)) {

    __m128i ctr1 = _mm_add_epi64(ctr0, add);

    // ks = enc(iv)
    zipaes.aes_double(ctr0, ctr1, ks0, ks1);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    ctr0 = _mm_add_epi64(ctr1, add);
  }
  
  // last block
  while (m_p < (long long)(_mlen)) {

    // ks = enc(iv)
    zipaes.aes(ctr0, ks0);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    //ctr0 = _mm_add_epi64(ctr0, add);
  }




}









// experiment
int parity(__m128i in, __m128i mask) {

  __m128i maskedIn = _mm_and_si128(in, mask);
  unsigned char maskedInB[16];
  _mm_store_si128((__m128i*)(maskedInB), maskedIn);

  int wt = 0;
  for (int j = 0; j < 16; j++) {
    wt += _mm_popcnt_u32(maskedInB[j]);
  }
  return (wt % 2);
}
double experimentDifferentialLinear(void) {

  __m128i ctr0 = _mm_setzero_si128();

  unsigned char diffB[16] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xB4, 0x00, 0x00, 0x00, 0x00 };
  __m128i diff = _mm_load_si128((const __m128i*)(diffB));

  unsigned char maskB[16] = { 0x00, 0x00, 0x00, 0x00, 0x32, 0xAB, 0x66, 0x98, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
  __m128i mask = _mm_load_si128((const __m128i*)(maskB));

  int N = (1 << 30);


  int numL = 0;
  int numR = 0;
  int num = 0;
  for (int i = 0; i < N; i++) {

    __m128i in0 = ctr0;
    __m128i in1 = _mm_xor_si128(in0, diff);

    __m128i out0L, out1L;
    __m128i out0R, out1R;

    zipaes.aesL(in0, out0L, 3);
    zipaes.aesL(in1, out1L, 3);

    zipaes.aesR(in0, out0R, 3);
    zipaes.aesR(in1, out1R, 3);

    __m128i out0 = _mm_xor_si128(out0L, out0R);
    __m128i out1 = _mm_xor_si128(out1L, out1R);

    __m128i outDiffL = _mm_xor_si128(out0L, out1L);
    __m128i outDiffR = _mm_xor_si128(out0R, out1R);
    __m128i outDiff = _mm_xor_si128(out0, out1);

    int pL = parity(outDiffL, mask);
    int pR = parity(outDiffR, mask);
    int p = parity(outDiff, mask);

    if (pL == 0) numL++;
    else numL--;

    if (pR == 0) numR++;
    else numR--;

    if (p == 0) num++;
    else num--;

    ctr0 = out0;
  }


  printf("3 + 3 rounds \n");
  printf("input difference : ");
  zipaes.dispXMM(diff);
  printf("output mask      : ");
  zipaes.dispXMM(mask);
  cout << "use 2^" << log2(N) << " pairs" << endl;
  cout << endl;


  double corL = (double)(numL) / (double)(N);
  double corR = (double)(numR) / (double)(N);
  double cor = (double)(num) / (double)(N);

  cout << "autocorrelation (left)  : " << abs(corL) << "\t" << log2(abs(corL)) << endl;
  cout << "autocorrelation (right) : " << abs(corR) << "\t" << log2(abs(corR)) << endl;
  cout << "autocorrelation         : " << abs(cor) << "\t" << log2(abs(cor)) << endl;

  return cor;
}
double experimentDifferentialAndLinear(bool rightGuess) {

  __m128i ctr0 = _mm_setzero_si128();

  // difference of the left side
  unsigned char inDiffB[16] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xB4, 0x00, 0x00, 0x00, 0x00 };
  __m128i inDiff = _mm_load_si128((const __m128i*)(inDiffB));

  // output linera mask
  unsigned char outMaskB[16] = { 0x00, 0x00, 0x00, 0x00, 0xCC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
  __m128i outMask = _mm_load_si128((const __m128i*)(outMaskB));

  // input mask of the right side
  unsigned char inMaskB[16] = { 0x00, 0x00, 0x00, 0x00, 0x72, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
  __m128i inMask = _mm_load_si128((const __m128i*)(inMaskB));

  //
  __m128i guess0 = zipaes.w128[0];
  __m128i guess6 = zipaes.w128[6];
  __m128i guess7 = zipaes.w128[7];
  __m128i guess8 = zipaes.w128[8];

  if (!rightGuess) {
    srand(time(NULL));
    unsigned char random[64];
    for (int i = 0; i < 64; i++) {
      random[i] = rand() & 0xFF;
    }

    guess0 = _mm_load_si128((const __m128i*)(random));
    guess6 = _mm_load_si128((const __m128i*)(random + 16));
    guess7 = _mm_load_si128((const __m128i*)(random + 32));
    guess8 = _mm_load_si128((const __m128i*)(random + 48));
  }



  //
  int N = (1 << 30);

  double numL = 0;
  double numR = 0;
  double num = 0;
  for (int i = 0; i < N; i++) {

    // prepare a pair satisfying differential
    __m128i in0 = ctr0;
    __m128i in1 = _mm_aesdeclast_si128(_mm_aesenclast_si128(_mm_xor_si128(in0, guess0), _mm_aesimc_si128(inDiff)), guess0);

    // 
    __m128i in0R, in1R;

    //
    in0R = _mm_xor_si128(in0, guess6);
    in0R = _mm_aesimc_si128(in0R);
    in0R = _mm_aesdec_si128(in0R, guess7);
    in0R = _mm_aesdec_si128(in0R, guess8);

    in1R = _mm_xor_si128(in1, guess6);
    in1R = _mm_aesimc_si128(in1R);
    in1R = _mm_aesdec_si128(in1R, guess7);
    in1R = _mm_aesdec_si128(in1R, guess8);

    int p0in = parity(in0R, inMask);
    int p1in = parity(in1R, inMask);


    // debug for differential
    __m128i out0L, out1L;
    zipaes.aesL(in0, out0L, 3);
    zipaes.aesL(in1, out1L, 3);
    int pL = parity(_mm_xor_si128(out0L, out1L), outMask);
    if (pL == 0) numL++;
    else numL--;



    // debug for linear
    __m128i out0R, out1R;
    zipaes.aesR(in0, out0R, 3);
    zipaes.aesR(in1, out1R, 3);
    int pR = p0in ^ p1in ^ parity(_mm_xor_si128(out0R, out1R), outMask);
    if (pR == 0) numR++;
    else numR--;



    // output
    __m128i out0, out1;
    zipaes.aes(in0, out0, 3);
    zipaes.aes(in1, out1, 3);
    int q = p0in ^ p1in ^ parity(_mm_xor_si128(out0, out1), outMask);
    if (q == 0) num++;
    else num--;


    //
    ctr0 = out0;

  }



  printf("(1 + 2) + (2 + 1) rounds \n");
  printf("input difference (left) : ");
  zipaes.dispXMM(inDiff);
  printf("input mask (right)      : ");
  zipaes.dispXMM(inMask);
  printf("output mask             : ");
  zipaes.dispXMM(outMask);
  cout << "use 2^" << log2(N) << " pairs" << endl;
  cout << endl;

  double corL = (double)(numL) / (double)(N);
  double corR = (double)(numR) / (double)(N);
  double cor = (double)(num) / (double)(N);

  cout << "autocorrelation for the left side      : " << abs(corL) << "\t" << log2(abs(corL)) << endl;
  cout << "squared correlation for the right side : " << abs(corR) << "\t" << log2(abs(corR)) << endl;
  cout << "total correlation for diff-and-lin     : " << abs(cor) << "\t" << log2(abs(cor)) << endl;
  




  return cor;
}

